/******************************************************************\
*                                                                  *
* KISSFILT.C  --  Filters a KISS log file by callsign              *
*                                                                  *
* Copyright 1991,1992 Paul Williamson, KB5MU                       *
*                                                                  *
* This program removes unwanted frames from a KISS log file.       *
* This reduces the size of the file for archival or transmission.  *
*                                                                  *
* When invoked without arguments, it produces a usage message.     *
*                                                                  *
* If invoked with two filename arguments, it takes KISS input      *
* from the first file and delivers filtered KISS output to the     *
* second file, and prompts the user for what action to take for    *
* each source:destination callsign pair.  If invoked with three    *
* filename arguments, the third filename contains the list of      *
* source:destination callsign pairs to be retained.                *
*                                                                  *
\******************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <conio.h>
#include "regexp.h"

#define	BUFLEN 600		/* big enough for all bytes to be transposed */

		/* Special characters for KISS */

#define	FEND		0300	/* Frame End */
#define	FESC		0333	/* Frame Escape */
#define	TFEND		0334	/* Transposed frame end */
#define	TFESC		0335	/* Transposed frame escape */
 
#define	ACCEPT	1		/* keep this frame */
#define	REJECT	0		/* discard this frame */

#define	TOP		1		/* add this spec at the top of the list */
#define	BOTTOM	0		/* add this spec at the bottom of the list */

enum action {_keep, _discard, _ask, _else, _info_only, _pid, _fileno };

/* Filtering by protocol ID */

enum pid_action_type {	PID_DISCARD=0,
								PID_KEEP,  };

enum pid_action_type pid_action[256];	/* Actions on PIDs. Init'd to 0. */

int	pid_filter = 0;		/* Flag; if nonzero, we filter by PID */

/* Filtering by file number in PB frames */

int fileno_filter = 0;		/* Flag; if nonzero, we filter by fileno */

unsigned long keep_fileno;

/* Data for filtering by type of AX.25 frame */

#define	NUM_FTYPES	11
enum 
frame_type	{I, UI, RR, RNR, REJ, SABM, DISC, DM, UA, FRMR, UNK};
char	*ftype_name[] =
						{"I", "UI", "RR", "RNR", "REJ",
						 "SABM", "DISC", "DM", "UA", "FRMR", "UNK"};

int	line_no = 0;
enum action	else_action = _ask;			/*  default behavior */
int	info_only = 0;							/*  by default, don't filter on type */
char	whitespace[] = " \t\n";
char	*current_re;

		/* Function Prototypes */

int	cdecl main(int argc, char **argv);
int	get_answer(char *acceptable_answers);
int	get_frame(unsigned char *buf, FILE *infile);
char	*getcall(unsigned char *framebuf, char *callbuf);
void	read_filter_file(FILE *filtfile);
void	do_action(enum action action);
void	do_pid_action(void);
void	do_fileno_action(void);
void	add_spec(enum action action, int where);
void	allocate_spec(enum action action, regexp *src, regexp *dest, int where);
void	dump_specs(void);
int  accept_frame(char *src, char *dest, enum frame_type ftype,
								int pid, unsigned long fileno);
int	callsign_match(char *call, regexp *re);
int	act_on(enum action fate, char *src, char *dest);
void	filt_frame(unsigned char *buf, int len, FILE *outfile);
void	put_frame(unsigned char *buf, int len, FILE *outfile);
enum	frame_type ftype_of(char *buf);
int	pid_of(char *buf);
unsigned long fileno_of(char *buf);

/* Regular Expression routines */
extern	regexp *regcomp(char *exp);
extern	int     regexec(regexp *prog, char *string);

/*******************\
*                   *
* The Main Program  *
*                   *
\*******************/

int cdecl main(int argc, char **argv)
{
FILE	*infile, *outfile, *filtfile = NULL;
int	status = 0;
unsigned char	buffer[BUFLEN];
int	len;
char	*ffn = NULL;
char	*p;

puts("KISSFILT - Filter a KISS log file by callsigns");
puts(" Version 1.3   2/17/92 Copyright 1991,1992 Paul Williamson KB5MU\n");

				/* Check the command line arguments */
if ((argc != 3) && (argc != 4))
	{
	puts("Usage:");
	puts("    kissfilt infile outfile [filterfile]");
	exit(1);
	}

						/* Open the input file */
if ((infile = fopen(argv[1], "rb")) == NULL)
	{
	fprintf(stderr, "Can't open file %s.\n", argv[1]);
	exit(1);
	}

						/* Now for the filter specification file */
if (argc == 4)
	ffn = argv[3];				/* filename was specified on command line */
else
	{
	printf("Would you like to use a filter file? [ynq] ");
	switch(get_answer("ynq"))
		{
		case 'y':
			printf("Yes\nFilename: ");
			if (fgets(buffer, BUFLEN, stdin) == NULL)
				{
			   printf("Aborted.\n");
				exit(1);
				}
			if ((p = strchr(buffer, '\n')) == NULL)
				{
				printf("Input too long.\n");
				exit(1);
				}
			*p = '\0';				/* trim the newline */
			ffn = buffer;
			break;

		case 'n':
			printf("No\n");
			break;

		case 'q':
			printf("Quit\n");
			exit(0);
			break;
		}	
	}

if (ffn)
	{
	if ((filtfile = fopen(ffn, "r")) == NULL)
		{
		fprintf(stderr, "Can't open file %s.\n", ffn);
		fclose(infile);
		exit(1);
		}
	else
		{
		fprintf(stderr, "Taking filter specification from %s.\n",
					_fullpath(NULL, strupr(ffn), 0));
		read_filter_file(filtfile);
// 		dump_specs();	//debug
		}
	}
	
						/* and finally the output file */
if ((outfile = fopen(argv[2], "wb")) == NULL)
	{
	fprintf(stderr, "Can't open file %s.\n", argv[2]);
	fclose(infile);
	exit(1);
	}

			/* Try to use big buffers to speed things up */
if (setvbuf(infile, (char *)NULL, _IOFBF, 8096) ||
	 setvbuf(outfile,(char *)NULL, _IOFBF, 8096) ||
		(filtfile && setvbuf(filtfile,(char *)NULL, _IOFBF, 8096)))
	puts("Couldn't use full buffering, so this may be slow.");

fprintf(stderr, "Scanning input KISS file...\n");

			/* Here's the main loop.  Get a frame, check if
			   we are to save it, and, if so, output it;
			   until we run out of frames to convert.  */

while ( ((len=get_frame(buffer, infile))!=0) &&
		!ferror(infile) &&
		!ferror(outfile))
	filt_frame(buffer, len, outfile);

			/* wrap up the input file */
if (ferror(infile))
	{
	perror("Error reading input file");
	status = 1;
	}

			/* wrap up the output files */
if (ferror(outfile))
	{
	perror("Error writing output file");
	status = 1;
	}

if (fclose(outfile))
	{
	perror("Error closing output file");
	status = 1;
	}

return status;		/* return value for batch files */
}

/***************************************************************\
*                                                               *
* get_frame                                                     *
*                                                               *
* This function gets one complete KISS frame of data from       *
* the specified input file, and places it in buf.  It returns   *
* the number of bytes placed in buf, or 0 if the frame was      *
* ill-formed or if no data is available.                        *
*                                                               *
* See the ARRL 6th Computer Networking Conference Proceedings,  *
* "The KISS TNC: A simple Host-to-TNC communications protocol"  *
* by Mike Chepponis, K3MC, and Phil Karn, KA9Q, for details of  *
* the KISS framing standard.                                    *
*                                                               *
\***************************************************************/

int get_frame(unsigned char *buf, FILE *infile)
{
unsigned char	*bufp;
int	chr;
int	count = 0;

newframe:

bufp = buf;

while (!ferror(infile) && !feof(infile) && ((chr = getc(infile)) != FEND))
	;				/* eat bytes up to first FEND */

while (!ferror(infile) && !feof(infile) && ((chr = getc(infile)) == FEND))
	;				/* eat as many FENDs as we find */

chr = getc(infile);		/* get past the nul frame-type byte */

while (!ferror(infile) && !feof(infile) && (chr != FEND) && bufp < buf+BUFLEN)
	{
	if (chr == FESC)
		{
		chr = getc(infile);
		if (chr == TFESC)
			*bufp++ = FESC;		/* put escaped character in buffer */
		else if (chr == TFEND)
			*bufp++ = FEND;
		else
			return 0;			/* ill-formed frame */
		}
	else
		*bufp++ = (unsigned char)chr;		/* put character in buffer */

	chr = getc(infile);
	}

if (bufp >= buf+BUFLEN)
	{
	fprintf(stderr, "Warning: KISS frame is too big!  Discarded.\n");
	goto newframe;
	}

return (bufp - buf);			/* return length of buffer */
}

/**************************************************************\
*                                                              *
* getcall                                                      *
*                                                              *
* This function translates a callsign from the AX.25 format    *
* in the frame buffer to ASCII format in the callsign buffer.  *
*                                                              *
* Returns the callsign buffer, for convenient use.             *
*                                                              *
\**************************************************************/

char *getcall(unsigned char *framebuf, char *callbuf)
{
int	i;
int	ssid;
char	*cbp = callbuf;

for (i=0; i<6; i++)		/* start by filling in the basic characters */
	if (*framebuf == 0x40)
		framebuf++;
	else
		*cbp++ = (char)((*framebuf++ >> 1) & 0x7F);	/* unshift to ASCII */

				/* tack on the SSID if it isn't zero */
if ((ssid = (*framebuf & 0x1E) >> 1) > 0)
	sprintf(cbp, "-%d", ssid);
else
	*cbp = '\0';

return callbuf;
}

/********************************************************************\
*                                                                    *
* read_filter_file                                                   *
*                                                                    *
* This routine scans in a file that defines what frames              *
* should be kept and which discarded.  The file contains             *
* lines in one of the following formats:                             *
*                                                                    *
* ; any line beginning with a semicolon is ignored                   *
* keep CALL1 CALL2                                                   *
* discard CALL1 CALL2                                                *
* ask CALL1 CALL2                                                    *
* else [ask|keep|discard]                                            *
*                                                                    *
* CALL1 and CALL2 may be regular expressions as implemented          *
* by Unix V8 regexp(3).  Any packet whose source and destination     *
* callsigns match CALL1 and CALL2, respectively, of a line that      *
* begins with "keep" will be copied to the output file.  Similarly,  *
* any packet whose callsigns match a "discard" spec will be removed  *
* from the file.  The user will be asked about any packet whose      *
* callsigns match an "ask" spec, and the response will be added to   *
* the table just like a "keep" or "discard" spec.                    *
*                                                                    *
* If a packet matches more than one line, the first line matched     *
* governs its fate.  If a packet matches no "keep", "discard", or    *
* "ask" specs, the "else" spec controls the resulting action.        *
* If there is no filter file, or if the filter file contains no      *
* "else" spec, the default behavior is "else ask".                   *
*                                                                    *
\********************************************************************/

struct keyword {
	enum action	action;
	char			*key;
	} keyword[] =
			{
			{ _keep,			"keep"		},
			{ _discard,		"discard"	},
			{ _ask,			"ask"			},
			{ _else,			"else"		},
			{ _info_only,	"info-only"	},
			{ _pid,			"pid"			},
			{ _fileno,		"PBPfile"	},
			{ 0,				NULL			}
			};

/* the specs from the file are kept in a linked list of these structs */
struct spec {
	struct spec			*next;
   regexp				*src_re;
	regexp				*dest_re;
	enum action			fate;
	} specs = { NULL },  *lastspec = &specs;

void read_filter_file(FILE *filtfile)
{
char	linebuf[80];				/*  the original line from the file */
char	*token;
struct keyword *kwp;

while (fgets(linebuf, 80, filtfile))
	{
	line_no++;
	if (strchr(linebuf, '\n') == NULL)		/*  check for too-long line */
		while (fgetc(filtfile) != '\n' && !feof(filtfile) && !ferror(filtfile))
			;		/*  just toss the extra linelength */

	if ((token=strtok(linebuf, whitespace)) != NULL	/*  get the first token */
			&& (*token != ';'))								/*  unless it's a comment. */
		{
		for (kwp = keyword; kwp->key; kwp++)		/*  search the keyword list */
			{
			if (stricmp(kwp->key, token) == 0)		/*  look for a match */
				{
				do_action(kwp->action);			/*  here it is; interpret the spec. */
				break;								/*  and stop searching. */
				}
			}
		if (kwp->key == NULL)			/*  if we reached the end without matching */
			fprintf(stderr, "Unknown keyword \"%s\", line %d ignored.\n",
						token, line_no);

		} /*  if there's a first token, not a comment */
	} /*  while */

if (ferror(filtfile))
	{
	perror("Error reading filter file");
	exit(1);
	}
}

/**************************************************************\
*                                                              *
* do_action                                                    *
*                                                              *
* This routine handles a single command from the filter file.  *
* Else commands are handled internally.  Other commands are    *
* passed to add_spec.                                          *
*                                                              *
\**************************************************************/

void do_action(enum action action)
{
char	*token;
struct keyword *kwp;

switch(action)
	{
	case _info_only:
		info_only++;
		break;

	case _pid:
		pid_filter++;
		do_pid_action();
		break;

	case _fileno:
		fileno_filter++;
		do_fileno_action();
		break;

	case _else:
		if ((token=strtok(NULL, whitespace)) != NULL) /*  get the next token */
			{
			for (kwp = keyword; kwp->key; kwp++)		/*  search the keyword list */
				{
				if (stricmp(kwp->key, token) == 0)		/*  look for a match */
					{
					switch(kwp->action)
						{
						case _keep:
						case _discard:
						case _ask:
							else_action = kwp->action;
/* 							printf("Else action is %s.\n",			// debug */
/* 										keyword[(int)else_action].key); */
							return;
						default:
							fprintf(stderr, "Else action \"%s\" is not valid in line %d.\n",
											token, line_no);
							break;
						}
					break;								/*  and stop searching. */
					}
				}
			if (kwp->key == NULL)			/*  if we reached the end without matching */
				fprintf(stderr, "Else action \"%s\" is not valid in line %d.\n",
							token, line_no);
	
			break;
			}
		else
			fprintf(stderr, "Else action missing in line %d.\n", line_no);
	
	default:
		add_spec(action, BOTTOM);
		break;
	}
}

/******************************************************\
*                                                      *
* do_pid_action                                        *
*                                                      *
* This routine handles a pid line in the filter file.  *
*                                                      *
\******************************************************/

void	do_pid_action(void)
{
char	*token;
int	pid_val;

while ((token=strtok(NULL, whitespace)) != NULL)	/* Get the next token */
	{
	if (sscanf(token, "%x", &pid_val) != 1)
		{
		fprintf(stderr, "PID value \"%s\" invalid in line %d.\n",
							token, line_no);
		continue;
		}
	else if ((pid_val < 0) || (pid_val > 255))
		{
		fprintf(stderr, "PID value \"%s\" out of range in line %d.\n",
							token, line_no);
		}
	else
		{
		pid_action[(int)pid_val] = PID_KEEP;
		}
   }

}

/*******************************************************\
*                                                       *
* do_fileno_action                                      *
*                                                       *
* This routine handles a file line in the filter file.  *
*                                                       *
\*******************************************************/

void	do_fileno_action(void)
{
char	*token;

if ((token=strtok(NULL, whitespace)) != NULL)	/* Get the next token */
	{
	if (sscanf(token, "%x", &keep_fileno) != 1)
		{
		fprintf(stderr, "FILE value \"%s\" invalid in line %d.\n",
							token, line_no);
		fileno_filter = 0;
		}
	}
else
	{
	fprintf(stderr, "FILE spec in line %d requires an argument.\n",
							line_no);
	fileno_filter = 0;
	}	
}

/**********************************************************\
*                                                          *
* add_spec                                                 *
*                                                          *
* This routine creates an entry for the filter spec list.  *
* This involves compiling each filter spec into a regexp.  *
*                                                          *
\**********************************************************/

void add_spec(enum action action, int where)
{
char	*token;
regexp	*src, *dest;

/*  printf("%s found in line %d.\n", keyword[(int)action].key, line_no);	//debug */

if ((token=strtok(NULL, whitespace)) != NULL) /*  get the next token */
	{
	strupr(token);			/*  callsigns are always uppercase! */
/* 	printf("Src: %s\t", token);	//debug */
	current_re = token;
	if ((src = regcomp(token)) == NULL)
		return;				/*  don't record a spec that didn't compile! */
	}
else
	{
	fprintf(stderr, "Callsigns missing in line %d.\n", line_no);
	return;
	}

if ((token=strtok(NULL, whitespace)) != NULL) /*  get the next token */
	{
	strupr(token);			/*  callsigns are always uppercase! */
/* 	printf("Dest: %s\n", token);	//debug */
	current_re = token;
	if((dest = regcomp(token)) == NULL)
		return;				/*  don't record a spec that didn't compile! */
	}
else
	{
	fprintf(stderr, "Second callsign missing in line %d.\n", line_no);
	return;
	}

allocate_spec(action, src, dest, where);
}

/***********************************************************\
*                                                           *
* allocate_spec                                             *
*                                                           *
* This routine adds a filter spec to the filter spec list.  *
* It allocates memory for the spec, fills it in, and links  *
* it at the end of the list.                                *
*                                                           *
\***********************************************************/

void allocate_spec(enum action action, regexp *src, regexp *dest, int where)
{
struct spec	*newspec;

if ((newspec = (struct spec *)malloc(sizeof *newspec)) == NULL)
	{
	fprintf(stderr, "Out of memory for filter specs in line %d.\n", line_no);
	exit(1);
	}
newspec->src_re = src;								/*  fill it up */
newspec->dest_re = dest;
newspec->fate = action;

if (where == BOTTOM)
	{
	newspec->next = NULL;						/*  link it at end of the list */
	lastspec->next = newspec;
	lastspec = newspec;
	}
else
	{
	newspec->next = specs.next;				/*  link it at the top of the list */
	specs.next = newspec;
	}
}

/******************************************************\
*                                                      *
* dump_specs                                           *
*                                                      *
* This routine is for debug only.  It prints out some  *
* of the information in the filter spec list.          *
*                                                      *
\******************************************************/

#if 0
void dump_specs(void)
{
struct spec *p;
int	i = 0;

printf("Summary of filter spec:\n");
for (p=specs.next; p; p = p->next)
	printf("Action = %s, Src=%s, Dest=%s\n", keyword[(int)(p->fate)].key,
					p->src_re, p->dest_re);

do {
	printf("%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\n",
													pid_action[i],
													pid_action[i+1],
													pid_action[i+2],
													pid_action[i+3],
													pid_action[i+4],
													pid_action[i+5],
													pid_action[i+6],
													pid_action[i+7],
													pid_action[i+8],
													pid_action[i+9],
													pid_action[i+10],
													pid_action[i+11],
													pid_action[i+12],
													pid_action[i+13],
													pid_action[i+14],
													pid_action[i+15]);
	i += 16;
	} while (i < 256);

}
#endif

/***************************************************************\
*                                                               *
* accept_frame                                                  *
*                                                               *
* This routine decides what to do with a frame.                 *
* It first searches the frame spec list.  If it finds a match   *
* there, it returns the action specified by that entry.  If no  *
* match is found, it returns the default action.                *
*                                                               *
\***************************************************************/

int  accept_frame(char *src, char *dest, enum frame_type ftype,
								int pid, unsigned long fileno)
{
struct spec *sp;

if (info_only)				/* if info_only requested, reject all non-I frames */
	switch(ftype)
		{
		case I:
		case UI:
		case FRMR:		/* FRMR has some info.  So keep it. */
			break;
		default:
			return REJECT;
			break;
		}

if (pid_filter)
	{
   if ((ftype == I) || (ftype == UI))
		switch(pid_action[pid])
			{
			case PID_KEEP:
					break;

			default:
					return REJECT;
			}
   else
		return REJECT;
	}

if (fileno_filter)
	{
	if ((ftype != UI) || (fileno != keep_fileno))
		return REJECT;
	}

for (sp=specs.next; sp; sp = sp->next)		/*  search through the spec list */
	{
	if (callsign_match(src, sp->src_re) && callsign_match(dest, sp->dest_re))
		return act_on(sp->fate, src, dest);
	}

return act_on(else_action, src, dest);		/*  if not found, take default */
}

/*******************************************************\
*                                                       *
* callsign_match                                        *
*                                                       *
* This routine decides if a callsign matches a regexp.  *
*                                                       *
\*******************************************************/

int callsign_match(char *call, regexp *re)
{
/*  printf("Checking callsign %s against %p - ", call, (void *)re);	//debug */

if (regexec(re, call) && ((re->endp[0] - re->startp[0]) ==
					(signed int)strlen(call)))
	{
/* 	printf("matched.\n");		//debug */
	return 1;
	}
else
	{
/* 	printf("no match.\n");		// debug */
	return 0;
	}
}

/**********************************************************\
*                                                          *
* act_on                                                   *
* This routine translates an action into an ACCEPT or      *
* REJECT.  This is trivial except in the case when the     *
* action is _ask.  In that case, it asks the user what     *
* to do, and records the results in the filter spec list.  *
*                                                          *
\**********************************************************/

int act_on(enum action fate, char *src, char *dest)
{
enum action	answer;
regexp	*src_re, *dest_re;

switch (fate)
	{
	case _keep:
		return ACCEPT;
	case _discard:
		return REJECT;
	case _ask:
		printf("Keep or discard packets from %-9s to %-9s? [kdq] ", src, dest);
		switch(get_answer("kdq"))
			{
			case 'k':
				answer = _keep;
				printf("Keep\n");
				break;
			case 'd':
				answer = _discard;
				printf("Discard\n");
				break;
			case 'q':
				printf("Quit\n");
				exit(0);
				break;
			}

		if ((src_re = regcomp(current_re = src)) == NULL)
			{
			fprintf(stderr, "Internal error: callsign %s didn't compile.\n",src);
			exit(1);
			}
		if ((dest_re = regcomp(current_re = dest)) == NULL)
			{
			fprintf(stderr, "Internal error: callsign %s didn't compile.\n", dest);		
			exit(1);
			}
		allocate_spec(answer, src_re, dest_re, TOP);	/*  add at top of list */
		return (answer == _keep) ? ACCEPT : REJECT;

	default:
		{
		fprintf(stderr, "Internal error: bogus action in act_on.\n");
		exit(1);
		}
	}

return 0;	/*  not reached */
}

int get_answer(char *acceptable_answers)
{
int	key;

fflush(stdout);
do
	{
	key = getch();
	key = tolower(key);
	if (key == 0)
		{
		getch();							/* handle two-byte keystrokes */
		putchar('\a');
		}
	else if (strchr(acceptable_answers, key) == NULL)
		putchar('\a');					/* beep for wrong keystrokes */
	else
		return key;
	} while (1);
}

/************************************************************\
*                                                            *
* filt_frame                                                 *
*                                                            *
* This routine accepts a buffer of specified length.         *
* It extracts the source and destination callsigns from it.  *
* If accept_frame() says the source and destination calls    *
* are to be accepted, it puts the frame to outfile.          *
*                                                            *
\************************************************************/

void filt_frame(unsigned char *buf, int len, FILE *outfile)
{
char	dest[10], src[10];		/*  put this frame's callsigns here */
enum	frame_type ftype;
unsigned char pid;
unsigned long fileno;

getcall(buf, dest);				/*  translate the callsigns */
getcall(buf+7, src);
ftype = ftype_of(buf);			/*  and figure out what kind of frame it is */
pid = (unsigned char) pid_of(buf);

if (fileno_filter)
	fileno = fileno_of(buf);

/* printf("%s > %s : %s, %d bytes, pid %02x\n", src, dest, ftype_name[ftype], len, pid);// debug */

if (accept_frame(src, dest, ftype, pid, fileno))
	put_frame(buf, len, outfile);
}

/************************************************************\
*                                                            *
* put_frame                                                  *
*                                                            *
* This function gets one complete frame of data in a buffer  *
* of specified length, and writes it as a KISS frame to the  *
* output file.  See the article referenced above.            *
*                                                            *
\************************************************************/

void put_frame(unsigned char *buf, int len, FILE *outfile)
{
int	i;

putc(FEND, outfile);				/*  all frames begin with FESC */
putc(0, outfile);					/*  then a frame type byte, namely 0 */
for (i=0; i<len; i++)
	{
	switch (buf[i])				/*  translate for data transparency */
		{
		case FEND:
			putc(FESC, outfile);
			putc(TFEND, outfile);
			break;
		case FESC:
			putc(FESC, outfile);
			putc(TFESC, outfile);
			break;
		default:
			putc(buf[i], outfile);
			break;
		}
	}
putc(FEND, outfile);				/*  all frames end with FESC */
}

/*************************************************************\
*                                                             *
* regerror                                                    *
*                                                             *
* This routine is required by Henry Spencer's regexp code.    *
* It is called with an explanatory message whenever an error  *
* is detected in his code.                                    *
*                                                             *
\*************************************************************/

void regerror(char *note)
{
fprintf(stderr, "Callsign spec \"%s\" in line %d is bad: %s\n",
			current_re, line_no, note);
}

/********************************************************\
*                                                        *
* ftype_of                                               *
*                                                        *
* This routine returns the frame type of a given frame.  *
*                                                        *
\********************************************************/

enum frame_type ftype_of(char *buf)
{
char	*bufp;
char	ctl;
static const enum frame_type sframes[] = 
			{RR, RNR, REJ, UNK, SABM, DISC, DM, UA, FRMR};


			/* skip the digipeater calls, if any */
if ((buf[13] & 1) == 0)		/* if there are digipeaters, */
	for (bufp = buf+14; !(*(bufp-1) & 1) && bufp < buf+70; bufp+=7)
		;
else
	bufp = buf+14;				/* point to CTL byte */

ctl = *bufp;			/* translate the ctl byte */

	/* now handle each frame type separately */

if ((ctl & 1) == 0)		/* if it's an I frame */
	return I;

else if ((ctl & 3) == 1)	/* if it's an S frame */
	return sframes[(ctl & 0x0C) >> 2];

else if ((ctl & 0xEF) == 0x03)	/* if it's a UI frame */
	return UI;

else				/* it's some other U frame */
	{
	switch (ctl & 0xEF)
		{
		case 0x2F:		/* SABM frame */
			return SABM;
			break;
		case 0x43:		/* DISC frame */
			return DISC;
			break;
		case 0x0F:		/* DM frame */
			return DM;
			break;
		case 0x63:		/* UA frame */
			return UA;
			break;
		case 0x87:		/* FRMR frame */
			return FRMR;
			break;
		default:
			return UNK;
			break;
		}
	}
}

/********************************************************\
*                                                        *
* pid_of                                                 *
*                                                        *
* This routine returns the Protocol ID of a given frame. *
* If the frame has no protocol ID, it returns -1.        *
*                                                        *
\********************************************************/

int pid_of(char *buf)
{
char	*bufp;
char	ctl;

			/* skip the digipeater calls, if any */
if ((buf[13] & 1) == 0)		/* if there are digipeaters, */
	for (bufp = buf+14; !(*(bufp-1) & 1) && bufp < buf+70; bufp+=7)
		;
else
	bufp = buf+14;				/* point to CTL byte */

ctl = *bufp;			/* translate the ctl byte */

	/* now handle each frame type separately */

if (((ctl & 1) == 0)	 ||		/* if it's an I frame, or */
    ((ctl & 0xEF) == 0x03))	/* if it's a UI frame */
	{
	return *++bufp;				/* return the protocol ID byte */
	}
else
	return -1;
}

/********************************************************\
*                                                        *
* fileno_of                                              *
*                                                        *
* This routine returns the file number of a given frame, *
* if the frame seems to be a Pacsat Broadcast Protocol   *
* broadcast frame.  If not, it returns 0.                *
*                                                        *
\********************************************************/

unsigned long fileno_of(char *buf)
{
char	*bufp;
char	ctl;

			/* skip the digipeater calls, if any */
if ((buf[13] & 1) == 0)		/* if there are digipeaters, */
	for (bufp = buf+14; !(*(bufp-1) & 1) && bufp < buf+70; bufp+=7)
		;
else
	bufp = buf+14;				/* point to CTL byte */

ctl = *bufp;			/* translate the ctl byte */

	/* now handle each frame type separately */

if (((ctl & 0xEF) == 0x03) &&
		((*++bufp & 0xFF) == 0xBB))		/* if it's a UI frame with PID 0xBB */
	{
	return *(unsigned long *)(bufp+2);	/* Return what might be file number */
	}
else
	return 0;
}

